import royal

from royal.exceptions import NotFound
from voluptuous import Schema, Required, Coerce, All, Range

from example.model import Photo


class Collection(royal.Collection):

    index_schema = Schema({
        Required('offset', default=0): All(Coerce(int), Range(min=0)),
        Required('limit', default=20): All(Coerce(int), Range(min=1, max=50)),
    })

    def index(self, params):
        offset = params['offset']
        limit = params['limit']
        cursor = Photo.get_newests(self.root.db, offset, limit)
        query = dict(offset=offset, limit=limit)
        return royal.PaginatedResult(self, cursor, Item, query,
                                     cursor.count())


class Item(royal.Item):

    def __init__(self, key, parent, document=None):
        super(Item, self).__init__(key, parent)
        self.document = document

    def load_document(self):
        if self.document is None:
            self.document = Photo.get_by_id(self.root.db, self.__name__)

        if self.document is None:
            raise NotFound(self)

        return self.document

    def show(self, params=None):
        result = self.load_document()
        author_username = result['author']
        result['href'] = self.url()
        result['author'] = {
            'href': self.root['users'][author_username].url()
        }
        return result


@royal.renderer_adapter(Item)
def adapt_item(item, request):
    return item.show()





def includeme(config):
    from .root import Root
    config.set_root_factory(Root)

    config.add_resource('users')
    config.add_resource('users.photos')
    config.add_resource('photos')
    config.scan(__name__)



import pkg_resources

from royal.tests.functional import TestBase


class Test(TestBase):

    def _add_users(self):
        from example.model import User
        for i in range(50):
            name = u'user%s' % i
            user = User.create(self.db, name, name + '@email.com')
            setattr(self, 'user%s' % i, user)
        self.addCleanup(self._delete_users)

    def _delete_users(self):
        for i in range(50):
            user = getattr(self, 'user%s' % i)
            user.delete()

    def _add_single_user(self, name, email):
        from example.model import User
        user = User.create(self.db, unicode(name), unicode(email))
        self.addCleanup(user.delete)
        return user

    def test_root(self):
        response = self.app.get('/')
        result = response.json
        self.assertIn('users', result)
        self.assertIn('photos', result)
        self.assertEqual('http://localhost/users/', result['users']['href'])
        self.assertEqual('http://localhost/photos/', result['photos']['href'])

    def test_users_index(self):
        self._add_users()
        response = self.app.get('/users?offset=20&limit=20')

        result = response.json

        url_first = 'http://localhost/users/?limit=20&offset=0'
        url_previous = 'http://localhost/users/?limit=20&offset=0'
        url_next = 'http://localhost/users/?limit=20&offset=40'
        url_href = 'http://localhost/users/?limit=20&offset=20'

        self.assertEqual(url_first, result['first'])
        self.assertEqual(url_next, result['next'])
        self.assertEqual(url_previous, result['previous'])
        self.assertEqual(url_href, result['href'])

        self.assertEqual(20, len(result['users']))

        for user in result['users']:
            username = user['username']
            self.assertIn('_id', user)
            self.assertIn('href', user)
            self.assertIn('email', user)
            self.assertIn('photos', user)
            self.assertEqual('http://localhost/users/%s/photos/' % username,
                             user['photos']['href'])

    def test_post_photo(self):
        from example.model import Photo
        image_gif = pkg_resources.resource_stream('royal.tests.functional',
                                                  'image.gif')
        result = self.app.post(
            '/users/hadrien/photos',
            upload_files=[(u'image', u'image.gif', image_gif.read())]
        )
        photo = Photo.get_one(self.db, author='hadrien')
        location = 'http://localhost/photos/%s/' % str(photo._id)

        headers = result.headers
        self.assertIn('Content-Type', headers)
        self.assertIn('Location', headers)

        self.assertEqual('application/json; charset=UTF-8',
                         headers['Content-Type'])
        self.assertEqual(location,
                         headers['Location'])

        expected = {
            'href': location,
            '_id': str(photo._id),
            'author': {
                'href': 'http://localhost/users/hadrien/'
            }
        }
        self.assertEqual(expected, result.json)

    def test_replace_user(self):
        image_gif = pkg_resources.resource_stream('royal.tests.functional',
                                                  'image.gif')
        user = self._add_single_user('franky', 'franky@email.com')
        result = self.app.post(
            '/users/franky/photos',
            upload_files=[(u'image', u'image.gif', image_gif.read())]
        )
        result = self.app.put('/users/franky/', {
            'username': 'micheal',
            'email': 'micheal@email.com',
        })
        expected = {
            'href': 'http://localhost/users/micheal/',
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_id': str(user['_id']),
            'photos': {
                'href': 'http://localhost/users/micheal/photos/'
            },
        }
        self.assertEqual(expected, result.json)
        result = self.app.get('/users/micheal/photos/')
        self.assertEqual(1, len(result.json['photos']))

    def test_replace_user_with_hack1(self):
        user = self._add_single_user('franky', 'franky@email.com')
        result = self.app.post('/users/franky/', {
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_method': 'put',
        })
        expected = {
            'href': 'http://localhost/users/micheal/',
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_id': str(user['_id']),
            'photos': {
                'href': 'http://localhost/users/micheal/photos/'
            },
        }
        self.assertEqual(expected, result.json)

    def test_replace_user_with_request_param_hack(self):
        user = self._add_single_user('franky', 'franky@email.com')
        result = self.app.post('/users/franky/', {
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_method': 'put',
        })
        expected = {
            'href': 'http://localhost/users/micheal/',
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_id': str(user['_id']),
            'photos': {
                'href': 'http://localhost/users/micheal/photos/'
            },
        }
        self.assertEqual(expected, result.json)

    def test_replace_user_with_with_header_hack(self):
        user = self._add_single_user('franky', 'franky@email.com')
        headers = {'X-HTTP-Method-Override': 'PUT'}
        result = self.app.post(
            '/users/franky/',
            {
                'username': 'micheal',
                'email': 'micheal@email.com',
                '_method': 'put',
            },
            headers=headers
        )
        expected = {
            'href': 'http://localhost/users/micheal/',
            'username': 'micheal',
            'email': 'micheal@email.com',
            '_id': str(user['_id']),
            'photos': {
                'href': 'http://localhost/users/micheal/photos/'
            },
        }
        self.assertEqual(expected, result.json)



from zope.interface import implementer

from pyramid.decorator import reify
from pyramid.traversal import find_root

from royal import exceptions
from royal.interfaces import (
    IBase,
    ICollection,
    IItem,
    IRoot,
)


def includeme(config):
    pass


@implementer(IBase)
class Base(object):

    def __init__(self, name, parent):
        self.__name__ = unicode(name)
        self.__parent__ = parent
        if not hasattr(self, 'children'):
            self.children = {}

    def __getitem__(self, key):
        self.on_traversing(key)
        return self.children[key](key, self)

    def _not_allowed(self, name):
        raise exceptions.MethodNotAllowed(self, name)

    @property
    def index(self):
        self._not_allowed('index')

    @property
    def create(self):
        self._not_allowed('create')

    @property
    def show(self):
        self._not_allowed('show')

    @property
    def replace(self):
        self._not_allowed('replace')

    @property
    def update(self):
        self._not_allowed('update')

    @property
    def delete(self):
        self._not_allowed('delete')

    @reify
    def root(self):
        return find_root(self)

    def resource_url(self, resource, **query_params):
        kw = {'query': query_params}
        return self.root.request.resource_url(resource, **kw)

    def url(self, **query_params):
        return self.resource_url(self, **query_params)

    @property
    def parent(self):
        return self.__parent__

    @property
    def name(self):
        return self.__name__

    @property
    def links(self):
        _links = {name: {'href': cls(name, self).url()}
                  for name, cls in self.children.iteritems()}
        _links['href'] = self.url()
        return _links

    def on_traversing(self, key):
        pass


@implementer(IRoot)
class Root(Base):
    children = {}
    request = None

    def __init__(self, request):
        super(Root, self).__init__('', None)
        self.request = request

    def show(self, params):
        return self.links


@implementer(IItem)
class Item(Base):
    pass


@implementer(ICollection)
class Collection(Base):

    def __getitem__(self, key):
        self.on_traversing(key)
        if hasattr(self, 'item_cls'):
            return self.item_cls(key, self)
        return self.children[key](key, self)



import json
from decimal import Decimal

import venusian
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPException,
    HTTPUnsupportedMediaType,
)

from webob.multidict import NoVars, MultiDict
from zope.interface import implementer

from royal.interfaces import IDeserializer


def includeme(config):
    utility = Utility()
    config.registry.registerUtility(utility)
    config.add_directive('add_deserializer', add_deserializer)
    config.add_request_method(utility.deserialize_request_body,
                              'deserialized_body', reify=True)
    config.scan(__name__)


@implementer(IDeserializer)
class Utility(object):

    def __init__(self):
        self.accept_headers = MultiDict()
        self.deserializers = {}

    def add_deserializer(self, content_type, callable):
        self.deserializers[content_type] = callable
        self.accept_headers.add('Accept', content_type)

    def deserialize_request_body(self, request):
        content_type = request.content_type
        deserialized = None
        request.response.headers.extend(self.accept_headers)

        if content_type in self.deserializers:
            try:
                deserialized = self.deserializers[content_type](request)
            except HTTPException as http_exc:
                http_exc.headers.extend(self.accept_headers)
                raise
        else:
            raise HTTPUnsupportedMediaType(content_type,
                                           headers=self.accept_headers)

        return deserialized


def add_deserializer(config, content_type, deserializer_func):

    def callback():
        utility = config.registry.getUtility(IDeserializer)
        utility.add_deserializer(content_type, deserializer_func)

    intr = config.introspectable(
        category_name='Request body deserializers',
        discriminator=content_type,
        title=content_type,
        type_name='',
    )
    intr['deserializer'] = deserializer_func

    config.action(content_type, callback, introspectables=(intr, ))


class deserializer_config(object):

    def __init__(self, content_type):
        self.content_type = content_type

    def __call__(self, callable):

        def callback(context, name, callable):
            config = context.config.with_package(info.module)
            config.add_deserializer(self.content_type, callable, **settings)

        info = venusian.attach(callable, callback)
        settings = {'_info': info.codeinfo}

        return callable


@deserializer_config('application/json')
def deserialize_json(request):
    fp = request.body_file_seekable
    fp.seek(0)
    try:
        return json.load(fp, encoding=request.charset, parse_float=Decimal)
    except (ValueError, UnicodeDecodeError):
        raise HTTPBadRequest('invalid json body')


@deserializer_config('application/x-www-form-urlencoded')
@deserializer_config('multipart/form-data')
def deserialiaze_form_urlencoded(request):
    if isinstance(request.POST, NoVars):
        raise HTTPBadRequest('invalid %s body' % request.content_type)
    return request.POST.mixed()



import unittest
from pyramid.testing import DummyRequest


class Test(unittest.TestCase):

    def check_assertions(self, cls, name, parent=None):
        b = cls(name, parent)
        if parent is None:
            root = b
        else:
            root = parent.root
        self.assertEqual(b.__name__, name)
        self.assertEqual(parent, b.__parent__)
        self.assertIs(root, b.root)
        return b

    def test_base(self):
        from royal.resource import Base
        self.check_assertions(Base, 'name')

    def test_root(self):
        from royal.resource import Root
        request = DummyRequest()
        root = Root(request)

        self.assertEqual(request, root.request)
        self.assertIsNone(root.__parent__)
        self.assertEqual('', root.__name__)
        self.assertEqual(root, root.root)

    def test_collection(self):
        from royal.resource import Collection, Root
        from royal.exceptions import MethodNotAllowed
        request = DummyRequest()
        root = Root(request)
        collection = self.check_assertions(Collection, 'users', root)
        with self.assertRaises(MethodNotAllowed):
            collection.index()
        with self.assertRaises(MethodNotAllowed):
            collection.create([])
        with self.assertRaises(MethodNotAllowed):
            collection.delete()

    def test_resource(self):
        from royal.resource import Collection, Item, Root
        from royal.exceptions import MethodNotAllowed
        request = DummyRequest()
        root = Root(request)
        c = Collection('users', root)
        resource = self.check_assertions(Item, 'hadrien', parent=c)

        with self.assertRaises(MethodNotAllowed):
            resource.show()
        with self.assertRaises(MethodNotAllowed):
            resource.delete()
        with self.assertRaises(MethodNotAllowed):
            resource.replace()
        with self.assertRaises(MethodNotAllowed):
            resource.update()



import mimetypes

import royal

from voluptuous import Schema, Required, Coerce, All, Range

from example.model import Photo
from example.resources import photos


class Collection(royal.Collection):

    index_schema = Schema({
        Required('offset', default=0): All(Coerce(int), Range(min=0)),
        Required('limit', default=20): All(Coerce(int), Range(min=1, max=50)),
    })

    def index(self, query_params):
        offset = query_params['offset']
        limit = query_params['limit']
        cursor = Photo.get_newests(self.root.db, offset, limit,
                                   author=self.parent.name)
        documents = [photos.Item(str(doc._id), self.root, doc).show()
                     for doc in cursor]
        result = {
            'photos': documents,
            'href': self.url(offset=offset, limit=limit),
            'first': self.url(offset=0, limit=limit),
        }
        has_previous = offset > 0
        if has_previous:
            result['previous'] = self.url(offset=max(offset - limit, 0),
                                          limit=limit)
        has_next = len(documents) == limit
        if has_next:
            result['next'] = self.url(offset=offset + limit, limit=limit)
        return result

    def create(self, params):
        fs = params['image']
        author = unicode(self.parent.name)
        mime_type = mimetypes.guess_extension(fs.filename)
        doc = Photo.create(self.root.db, author, fs.file, mime_type)
        return photos.Item(str(doc._id), self.root['photos'], doc)



import os
import logging
import unittest

import webtest

from pyramid.config import Configurator
from pyramid.decorator import reify
from pyramid_mongokit import SingleDbConnection

log = logging.getLogger(__name__)


def setup_package():
    connection = new_connection()
    setup_example_fixtures(connection)
    connection.close()


def teardown_package():
    connection = new_connection()
    for name in connection.db.collection_names():
        if u'system.indexes' not in name:
            connection.db[name].remove()
    connection.close()


class TestBase(unittest.TestCase):

    maxDiff = None

    @reify
    def config(self):
        _config = Configurator(settings={})
        _config.include('example')
        self.addCleanup(delattr, self, 'config')
        return _config

    @reify
    def app(self):
        self.addCleanup(delattr, self, 'app')
        return webtest.TestApp(self.config.make_wsgi_app())

    @reify
    def db(self):
        self._db = new_connection()
        self.addCleanup(self._db.close)
        self.addCleanup(delattr, self, 'db')
        return self._db.get_db()


def new_connection():
    from example.model import User, Photo
    db_prefix = ''
    connection = SingleDbConnection(os.environ['MONGO_DB_NAME'], db_prefix,
                                    os.environ['MONGO_URI'])
    connection.register(User)
    connection.register(Photo)
    connection.generate_index(User)
    connection.generate_index(Photo)
    return connection


def setup_example_fixtures(connection):
    from example.model import User

    try:
        User.create(connection.db, u'hadrien', u'hadrien@ectobal.com')
    except Exception:
        log.exception('setup fixture failed')



from pyramid.events import NewRequest

from royal.renderer import renderer_adapter
from royal.resource import Root, Collection, Item
from royal.deserializer import deserializer_config


__all__ = [
    'Collection',
    'deserializer_config',
    'Item',
    'renderer_adapter',
    'Root',
]


def includeme(config):
    config.include('royal.renderer')
    config.include('royal.deserializer')
    config.include('royal.utility')
    config.include('royal.resource')
    config.include('royal.views')

    config.add_subscriber_predicate(
        'request_methods',
        RequestMethodEventPredicate
    )

    config.add_subscriber(
        override_request_method,
        NewRequest,
        request_methods=['POST'],
    )

    config.commit()


class RequestMethodEventPredicate(object):

    def __init__(self, methods, config):
        self.methods = methods
        self.phash = self.text

    def __call__(self, event):
        return event.request.method in self.methods

    def text(self):
        return 'request_method in %s' % self.methods


def override_request_method(event):
    methods = ['PUT', 'DELETE']
    override = (
        event.request.headers.get('X-HTTP-Method-Override') or
        event.request.POST.get('_method', '').upper()
    )
    if override in methods:
        event.request.method = override



import logging

from pyramid.interfaces import IRootFactory
from zope.interface import implementer

from royal import exceptions
from royal.interfaces import IResourceConfigurator

log = logging.getLogger(__name__)


def includeme(config):
    utility = config.registry.queryUtility(IResourceConfigurator)

    if utility is None:
        utility = ResourceConfigurator()
        config.registry.registerUtility(utility, name=u'')
        config.add_directive('add_resource', utility.add_resource)


class ResourceDefinition(object):
    def __init__(self, resource_path, collection_cls, item_cls, parent,
                 name):
        self.resource_path = resource_path
        self.collection_cls = collection_cls
        self.item_cls = item_cls
        self.parent = parent
        self.name = name

    def __repr__(self):
        return '<ResourceDefinition(%s)>' % self.resource_path


@implementer(IResourceConfigurator)
class ResourceConfigurator(object):
    """Internal configurator of royal resources.
    """

    intr_category_name = 'Royal resources'

    def __init__(self):
        self.definitions = {}
        self._ainfo = None

    def add_resource(self, config, resource_path):
        try:
            parent, name = resource_path.rsplit('.', 1)
        except ValueError:
            # no separator in resource name: it is a root resource
            parent = None
            name = resource_path

        module_name = '%s.%s' % (config.package_name,
                                 resource_path.replace('.', '_'))

        module = config.maybe_dotted(module_name)

        collection_cls = getattr(module, 'Collection', None)
        item_cls = getattr(module, 'Item', None)

        if not any((collection_cls, item_cls)):
            raise exceptions.InvalidConfig(
                'no collection or item class found in %s', module)

        if collection_cls:
            collection_cls.children = {}

        if item_cls:
            item_cls.children = {}

        discriminator = ('royal', resource_path)

        intr = config.introspectable(
            category_name=self.intr_category_name,
            discriminator=discriminator,
            title=resource_path,
            type_name='Resource',
        )
        intr['collection_cls'] = collection_cls
        intr['item_cls'] = item_cls
        if parent:
            intr.relate(self.intr_category_name, ('royal', parent))
        else:
            intr.relate('root factories', None)

        definition = ResourceDefinition(resource_path, collection_cls,
                                        item_cls, parent, name)

        self.definitions[resource_path] = definition

        # Actions must be taken from top-level resource to leaf after
        # root_factory has been set at order=0
        # let's use order parameter:
        order = 1 + resource_path.count('.')
        # 'app.user.message' would be at order 3
        # 'app.user' at 2
        # 'app' at 1

        config.action(discriminator, self.register_resource,
                      args=(config, resource_path), order=order,
                      introspectables=(intr,))

    def register_resource(self, config, resource_path):
        definition = self.definitions[resource_path]
        # get resource definition of parent
        parent_def = (self.definitions[definition.parent]
                      if definition.parent
                      else config.registry.queryUtility(IRootFactory))

        # get resource parent class
        parent_cls = (parent_def.item_cls
                      if hasattr(parent_def, 'item_cls')
                      else parent_def)

        parent_cls.children[definition.name] = (definition.collection_cls
                                                if definition.collection_cls
                                                else definition.item_cls)

        if all((definition.collection_cls, definition.item_cls)):
            definition.collection_cls.item_cls = definition.item_cls

        log.debug('add resource=%s parent_def=%s parent_cls=%s',
                  resource_path, parent_def, parent_cls)



import royal

from voluptuous import Schema, Required, Coerce, All, Range
from royal.exceptions import NotFound

from example.model import User, Photo


class Collection(royal.Collection):

    index_schema = Schema({
        Required('offset', default=0): All(Coerce(int), Range(min=0)),
        Required('limit', default=20): All(Coerce(int), Range(min=1, max=50)),
    })

    def index(self, params):
        offset = params['offset']
        limit = params['limit']
        cursor = User.get_newests(self.root.db, offset, limit)
        documents = [Item(user.username, self, user).show() for user in cursor]
        result = {
            'users': documents,
            'href': self.url(offset=offset, limit=limit),
            'first': self.url(offset=0, limit=limit),
        }
        has_previous = offset > 0
        if has_previous:
            result['previous'] = self.url(offset=max(offset - limit, 0),
                                          limit=limit)
        has_next = len(documents) == limit
        if has_next:
            result['next'] = self.url(offset=offset + limit, limit=limit)
        return result


class Item(royal.Item):

    def __init__(self, key, parent, document=None):
        super(Item, self).__init__(key, parent)
        self.document = document

    def load_document(self):
        if self.document is None:
            self.document = User.get_one(self.root.db, username=self.name)
        if self.document is None:
            raise NotFound(self)
        return self.document

    def on_traversing(self, key):
        self.load_document()

    def show(self, params=None):
        user = self.load_document()
        user['href'] = self.url()
        user['photos'] = {
            'href': self['photos'].url(),
        }
        return user

    def replace(self, params):
        new_username = params['username']
        new_email = params['email']
        user = User.replace(self.root.db, self.name,
                            new_username=new_username,
                            new_email=new_email)
        Photo.replace_author(self.root.db, self.name, new_username)
        new_item = Item(new_username, self.parent, user)
        user['href'] = new_item.url()
        user['photos'] = {
            'href': new_item['photos'].url()
        }
        return user



import logging
from collections import OrderedDict

import venusian
from pyramid.interfaces import IRendererFactory
from pyramid.renderers import JSON as pyramid_JSON
from zope.interface import implementer

from royal.interfaces import IAdaptableRenderer

log = logging.getLogger(__name__)


def includeme(config):
    config.add_renderer('royal', Factory)

    config.registry.registerUtility(JSON())

    config.add_directive('add_renderer_adapter', add_renderer_adapter)

    config.scan(__name__)


@implementer(IRendererFactory)
class Factory(object):
    """ Constructor: info will be an oect having the
    following attributes: name (the renderer name), package
    (the package that was 'current' at the time the
    renderer was registered), type (the renderer type
    name), registry (the current application registry) and
    settings (the deployment settings dictionary). """

    default_match = 'application/json'

    formatters = None

    def __init__(self, info):
        self.info = info

        self.formatters = OrderedDict([
            ('application/json',
             info.registry.getUtility(IAdaptableRenderer)(info)),
        ])

    def __call__(self, value, system):
        """ Call the renderer implementation with the value
        and the system value passed in as arguments and return
        the result (a string or unicode oect).  The value is
        the return value of a view.  The system value is a
        dictionary containing available system values
        # (e.g. view, context, and request). """
        request = system['request']
        format = request.accept.best_match(self.formatters.keys(),
                                           default_match=self.default_match)
        request.response.content_type = format
        return self.formatters[format](value, system)


@implementer(IAdaptableRenderer)
class JSON(pyramid_JSON):
    pass


def add_renderer_adapter(config, dotted_name, adapter):

    def callback():
        (config.registry.getUtility(IAdaptableRenderer)
                        .add_adapter(adapted, adapter))

    adapted = config.maybe_dotted(dotted_name)

    intr = config.introspectable(
        category_name='Renderer adapters',
        discriminator=adapted,
        title=adapted,
        type_name='',
    )
    intr['adapter'] = adapter

    config.action(adapted, callback, introspectables=(intr, ))


class renderer_adapter(object):
    """A decorator to add a rendered adapter"""

    def __init__(self, type_or_iface):
        self.type_or_iface = type_or_iface

    def __call__(self, callable):

        def callback(context, name, ob):
            config = context.config.with_package(info.module)
            config.add_renderer_adapter(self.type_or_iface, ob, **settings)

        info = venusian.attach(callable, callback)
        settings = {'_info': info.codeinfo}
        return callable


@renderer_adapter('datetime.date')
@renderer_adapter('datetime.datetime')
def adapt_datetime(o, request):
    return o.isoformat()


@renderer_adapter('decimal.Decimal')
def adapt_decimal(o, request):

    class _number_str(float):
        # kludge to have decimals correctly output as JSON numbers.
        # converting them to strings would cause extra quotes.

        def __init__(self, o):
            self.o = o

        def __repr__(self):
            return str(self.o)

    return _number_str(o)



import logging

from pyramid.view import view_defaults, view_config
from pyramid.httpexceptions import (
    HTTPCreated,
    HTTPConflict,
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPNotFound,
    HTTPMethodNotAllowed,
)


from royal import exceptions as exc
from royal import interfaces

log = logging.getLogger(__name__)


def includeme(config):
    config.scan(__name__)


class BaseView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request


@view_defaults(context=interfaces.ICollection, renderer='royal')
class CollectionView(BaseView):

    @view_config(request_method='GET', permission='index')
    def index(self):
        func = self.context.index
        query_params = self.request.GET.mixed()
        if hasattr(self.context, 'index_schema'):
            result = func(self.context.index_schema(query_params))
        else:
            result = func(query_params)

        return result

    @view_config(request_method='POST', permission='create')
    def create(self):
        func = self.context.create

        params = self.request.deserialized_body
        if hasattr(self.context, 'create_schema'):
            params = self.context.create_schema(params)

        item = func(params)

        if hasattr(item, 'url'):
            self.request.response.headers['Location'] = item.url()
            self.request.response.status_int = HTTPCreated.code

        return item


@view_defaults(context=interfaces.IItem, renderer='royal')
class ItemView(BaseView):

    @view_config(request_method='GET', permission='show')
    @view_config(request_method='GET', permission='show',
                 context=interfaces.IRoot)
    def show(self):
        func = self.context.show
        params = self.request.GET.mixed()
        return func(params)

    @view_config(request_method='PUT', permission='replace')
    def put(self):
        func = self.context.replace
        params = self.request.deserialized_body
        return func(params)

    @view_config(request_method='PATCH', permission='update')
    def patch(self):
        func = self.context.update
        params = self.request.deserialized_body
        func(params)
        return self.request.response

    @view_config(request_method='POST')
    def post(self):
        # XXX should we permit POST on Item
        raise exc.MethodNotAllowed(self, 'POST')


@view_config(context=interfaces.IItem, request_method='DELETE',
             permission='delete', renderer='royal')
@view_config(context=interfaces.ICollection, request_method='DELETE',
             permission='delete', renderer='royal')
def delete(context, request):
    context.delete()
    return {}


@view_config(context=exc.MethodNotAllowed, renderer='royal')
@view_config(context=interfaces.ICollection, renderer='royal')
@view_config(context=interfaces.IItem, renderer='royal')
@view_config(context=interfaces.IRoot, renderer='royal')
def not_allowed(context, request):
    request.response.status_int = HTTPMethodNotAllowed.code
    return {
        'error': 'method_not_allowed',
        'resource': request.resource_url(context.resource
                                         if hasattr(context, 'resource')
                                         else context),
        'http_method': request.method
    }


@view_config(context=exc.NotFound, renderer='royal')
def item_not_found(context, request):
    request.response.status_int = HTTPNotFound.code
    return {
        'error': 'not_found',
        'resource': request.resource_url(context.resource)
    }


@view_config(context=exc.Conflict, renderer='royal')
def conflict(context, request):
    request.response.status_int = HTTPConflict.code
    return {
        'error': 'already_exists',
        'resource': request.resource_url(context.resource)
    }


@view_config(context=exc.BadParameter, renderer='royal')
def bad_parameter(context, request):
    request.response.status_int = HTTPBadRequest.code
    return {
        'error': 'invalid_parameters',
        'resource': request.resource_url(context.resource),
        'message': '%s="%s"' % (context.name, context.value),
    }


@view_config(context=Exception, renderer='royal')
def exception(context, request):
    request.response.status_int = HTTPInternalServerError.code
    return {
        'error': 'unexpected_error',
        'message': unicode(context),
        'error_class': type(context).__name__,
    }



from bson import BSON
from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_config

from royal import deserializer_config


def main(global_config, **settings):  # pragma no cover
    config = Configurator(settings=settings)
    config.include(includeme)
    return config.make_wsgi_app()


def includeme(config):
    config.include('royal')
    config.include('example.model')
    config.include('example.resources')
    config.scan()


@deserializer_config('application/bson')
def deserialize_bson(request):
    return BSON(request.body.decode(request.charset)).decode()


@deserializer_config('multipart/form-data')
def deserialize_multipart_form_data(request):
    return request.POST.mixed()


@view_config(context='voluptuous.MultipleInvalid', renderer='royal')
def invalid_parameter(context, request):
    request.response.status_int = HTTPBadRequest.code
    return {
        'error': 'invalid_parameters',
        'message': unicode(context)
    }





class Base(Exception):
    """Base exception class for all royal exceptions."""


class InvalidConfig(Base):
    ""


class InternalError(Base):
    ""


class ResourceException(Base):

    def __init__(self, resource, *args):
        """
        :param resource: Resource the exception is related with.
        """
        super(Base, self).__init__(resource, *args)

    @property
    def resource(self):
        return self.args[0]


class Conflict(ResourceException):
    """Raised in case of conflict, i.e.: Duplicated resource."""


class NotFound(ResourceException):
    pass


class MethodNotAllowed(ResourceException):
    def __init__(self, resource, http_method):
        super(MethodNotAllowed, self).__init__(resource, http_method)

    @property
    def http_method(self):
        return self.args[1]


class BadParameter(ResourceException):

    def __init__(self, resource, name, value, msg=''):
        """Raised when bad parameters are provided

        :param name: Parameter name
        :param value: Parameter value
        """
        super(BadParameter, self).__init__(resource, name, value, msg)

    @property
    def name(self):
        return self.args[1]

    @property
    def value(self):
        return self.args[2]



from zope.interface import Attribute, Interface


class IBase(Interface):
    "Base for all context class."
    __name__ = Attribute('The __name__ attribute should be the name with '
                         'which a resource\'s parent refers to the resource '
                         'via __getitem__.')
    __parent__ = Attribute('Should be a reference to the resource\'s parent '
                           'resource instance in the tree')


class IRoot(IBase):
    "Root factory"
    request = Attribute('The request object.')
    links = Attribute('A mapping which gives relationship with other '
                      'resources.')


class IItem(IBase):

    def show(params):
        "GET /items/{id}"

    def update(params):
        "PUT /items/{id}"

    def patch(params):
        "PATCH /items/{id}"

    def delete():
        "DELETE /items/{id}"


class ICollection(IBase):

    def index(offset, limit, params):
        "GET /items?offset=0&limit=20"

    def create(params):
        "POST /items"

    def delete():
        "DELETE /items"


class IResourceConfigurator(Interface):

    def add_resource(dot_path):
        "Add a resource by its dot notation: 'apps.users' is apps/{id}/users"


class IDeserializer(Interface):

    def add_deserializer(content_type, callable):
        "Add a deserializer callable for a content type"

    def deserialize_request_body(request):
        "Deserialize request body according to its content type."


class IAdaptableRenderer(Interface):

    def add_adapter(type_or_iface, adapter):
        """ When an object of the type (or interface) ``type_or_iface`` fails
        to automatically encode using the serializer, the renderer will use
        the adapter ``adapter`` to convert it into a serializable object."""

    def __call__(info):
        """ Returns the encoded string with the content-type as the renderer
        defined (``application/json`` for a json renderer)."""
